#
# $Id: tofu.pl,v 1.10 2003/07/01 12:56:05 herbert Exp $
#
# Perl filtering file for the innd hooks.
#
# Filters out top-posted articles (TOFU) or those with excessive quoting.
# 

#############################################################################
# TOFU filter configuration: adapt these to your whishes:
#

# Only filter newsgroups matching this regular expression (leave empty to
# filter all)
#$tofu_filterng = "fhw\.";
$tofu_filterng = "";
# check the quote/lines ratio (default: off)
#$tofu_checkr = 1;
$tofu_checkr = 0;
# the maximum quote/lines ratio accepted (if $tofu_checkr != 0)
$tofu_bigqr = sprintf ("%1.2f", 1/1.5);
# the maximum number of continuous quote lines accepted (fullquote protection)
$tofu_bigqn = 30;
# Indent string, regexp to identify a quoted line
$tofu_indent = '[>]';
# debug messages to INN::syslog enabled (if $tofu_logging != 0)
#$tofu_logging = 1;
$tofu_logging = 0;
# log all filtered messages in /var/log/news/tofu.log (if $tofu_peeking != 0)
#$tofu_peeking = 1;
$tofu_peeking = 0;

#############################################################################
# ... no need to change anything from here to filter_art() ....
$tofu_bigqx = 10;
$tofu_trad = 1;

#############################################################################
# tofu_bigQuotes(): Finds big quotes (more than $n lines quoted) and deletes 
#                   all but $x lines of them.
# return:           empty string if nothing's been found, else a message 
#                   indicating excessive quoting
sub tofu_bigQuotes {
  my $L = shift;                # array of message lines
  my $V = shift;                # array with verbatim list
  my $k = 0;
  my $snipped = 0;
  my $nquote = 0;
  my $nlines = $#$L + 2;
  my $ratio = 1.0;
  my $rval = "";

  if ($tofu_logging != 0)
    {
      INN::syslog ("DEBUG", "Entering tofu_bigQuotes():");
    }

  # count the number of quoted lines
  for (my $i = 0; $i <= $#$L; $i++) 
    {
      if ($$L[$i] =~ /^ *$tofu_indent/) 
        { 
          $nquote++; 
        } 
    }
  
  $ratio = sprintf ("%1.2f", $nquote / $nlines);

  if ($tofu_logging != 0)
    {
      INN::syslog ("DEBUG", "$nquote lines of quote / $nlines lines: $ratio.");
    }

  if ($tofu_checkr != 0 && $ratio > $tofu_bigqr)
    {
      return "Excessive quotes, $ratio quote/lines," .
        " only $tofu_bigqr accepted [TOFU Filter].";
    }

  for (my $i = 0; $i <= $#$L; $i++) 
    {
      if ($$V[$i]) 
        { 
          $k = 0;
          next;
        }

      if ($$L[$i] =~ /^ *$tofu_indent/) 
        { 
          $k++; 
        } 
      else 
        { 
          if ($k > $tofu_bigqn) 
            {
              my $x = $k - $tofu_bigqx;
              $i -= $k;

              $$L[$i] = "[---=| Quote block shrinked by $0: " .
                "$x lines snipped |=---]\n";
              $snipped += $k;

              if ($tofu_logging != 0)
                {
                  INN::syslog ("DEBUG", 
                               "Long quote found: $k, snipped $snipped.");
                }

              $i++;
              splice (@$L, $i, $x-1);
              splice (@$V, $i, $x-1);

              $i++;
            }
          $k = 0;
        }
    }
  
  if ($snipped > 0)
    {
      $rval = "Too much quoted text in this message:" . 
        " $snipped lines [TOFU Filter].";
    }

  if ($tofu_logging != 0)
    {
      INN::syslog ("DEBUG", "Leaving tofu_bigQuotes(): \"$rval\".");
    }

  return $rval;
}

#############################################################################
# tofu_invalidQuoting(): Finds full quotes below the message text and deletes 
#                        them
# return:                Empty string if nothing's been found, else a message 
#                        indicating top posting
sub tofu_invalidQuoting {
  my $lines = shift;
  my @vrb = ();
  my @sig = ();

  if ($tofu_logging != 0)
    {
      INN::syslog ("DEBUG", "Entering tofu_invalidQuoting():");
    }

  # bloat this array if you want more internationalization:
  my @tofu = ('Original Message',
              'Ursprngliche Nachricht',
              'Ursprungliche Nachricht',
              'Mensagem original',
              'Ursprungligt meddelande',
              'Oorspronkelijk bericht',
              'Message d\'origine',
              'Weitergeleitete Nachricht / Forwarded Message');
  my $k = 0;                    # any text above?

  $verb = 0;
  for ($x = 0; $x < scalar (@$lines); $x++) 
    {
      if ($$lines[$x] =~ /^\s*  #v([+-])$/) 
        { 
          $verb = $1 eq '+' ? 1 : 0;
          $vrb[$x] = 1;
        } 
      else 
        {
          $vrb[$x] = $verb;
        }
    }

  # Remove signature:
  if (scalar(@$lines)) 
    { 
      my $sn = 0;
      for ($x = $#$lines; $x >= 0; $x--)
        {
          if ((!$vrb[$x]) && $$lines[$x] =~ /^ *-- $/)
            {
              if ($tofu_logging != 0)
                {
                  INN::syslog ("DEBUG", "Found sig: \"$$lines[$x + 1]\".");
                }
              @sig = (splice(@$lines, $x), @sig);
            }
        }
    }


 DONE: for ($x = 0; $x < scalar (@$lines); $x++) 
    { 
      if (!$vrb[$x]) 
        {
          foreach my $tmp (@tofu) 
            {
              if ($$lines[$x] =~ /^-+\s?$tmp\s?-+/) 
                { 
                  $x++;
                  if ($k) 
                    {
                      $tofu_trad = 0;
                    }
                  last DONE; 
                }
            }

          if ((!$k) && $$lines[$x] !~ /^\s*$/)
            {
              $k = 1;
            }
        }
    }

  # try to avoid false positives and only delete m$ style tofu
  # if there is text above
  if ($k) 
    {
      $j = scalar (@$lines)-$x;
      if ($tofu_logging != 0)
        {
          INN::syslog ("DEBUG", "Found MS TOFU: $j lines.");
        }
      splice (@$lines, $x);
    }

  # Nothing? Then try traditional TOFU (deleting M$ style TOFU is done
  # much more aggressively, so we won't need to search any more if we
  # did find some):
  if ($tofu_trad && (!$j) && !$vrb[$#$lines]) 
    {
      if (scalar (@$lines) && $$lines[$#$lines] =~ /^\s*$/) 
        { 
          unshift (@sig, pop (@$lines));
        }

      my $k;

      for (my $i = $#$lines; $i >= 0; $i--) 
        {
          if ($$lines[$i] =~ /^ *$tofu_indent/o) 
            {
              $j++;
              $k = $i;
            }
          elsif ($$lines[$i] !~ /^\s*$/) 
            {
              last;
            }
        }

      if ($j) 
        {
          if ($tofu_logging != 0)
            {
              INN::syslog ("DEBUG", "Found traditional TOFU: $j lines.");
            }
          splice (@$lines, $k);
        }
    }

  if ($tofu_logging != 0)
    {
      INN::syslog ("DEBUG", "Leaving tofu_invalidQuoting().");
    }

  # OK, if we found TOFU, we will leave a message that we were here...
  if ($j) 
    { 
      push(@$lines, "[---=| TOFU protection by $0: " .
           "$j lines snipped |=---]\n");
      if ($#sig)
        {
          push (@$lines, @sig);
        }
      return "Invalid quoting format" . 
        " (top posting is not accepted) [TOFU Filter].";
    }
  return tofu_bigQuotes ($lines, \@vrb); 
}

#############################################################################
# tofu_text2Array(): Converts text from a string to a (linewise) array
# return:            Array containing the text lines from the string
sub tofu_text2Array {
  my $text = shift;
  
  my @rval = split ("\n", "$text");
  for (my $i = 0; $i <= $#rval; $i++) 
    {
      $rval[$i] =~ s/[\n\r]*$/\n/;
    }
  return @rval;
}

#############################################################################
# filter_art(): Checks all articles innd receives from a peer.
# return:       "" to accept, any other string (should contain a meaningful
#               message) to reject. In case of rejection, the reason given 
#               by the return value will be logged.
#
# Debugging:    Set (one or both of) $tofu_logging and $tofu_peeking to 
#               values != 0 and look at /var/log/syslog and 
#               /var/log/news/tofu.log respectively.
#
# Features:     Comment-out some of the stuff below if you don't want it
#
sub tofu_runFilters {
  my ($maxgroups) = 10 ;
  my $rval = "";
  my @message = ();

  # Do not filter any articles to non-local newsgroups...
  if ($tofu_filterng && $hdr{'Newsgroups'} !~ /$tofu_filterng/)
    {
      if ($tofu_logging != 0)
        {
          INN::syslog ("DEBUG", "Skipping for NG \"$hdr{'Newsgroups'}\".");
        }
    }
  else
    {
      @message = tofu_text2Array ($hdr{"__BODY__"});
      $rval = tofu_invalidQuoting (\@message);
      if ($rval)
        {
          if ($tofu_logging != 0)
            {
              INN::syslog ("DEBUG", 
                           "Article rejected: NGs: \"$hdr{'Newsgroups'}\".");
            }
          if ($tofu_peeking != 0)
            {
              my $log = "Date: $hdr{'Date'}\n" .
                "Newsgroups: $hdr{'Newsgroups'}\n" .
                  "From: $hdr{'From'}\n" . 
                    "Subject: $hdr{'Subject'}\n\n" .
                      "$hdr{'__BODY__'}\n" .
                        "$rval\n\n";
              open (OUT, ">>/var/log/news/tofu.log");
              print OUT "$log\n";
              close (OUT);
            }
        }
    }

  return $rval;
}

1;
